(* variables utilisées pour lancer glucose et glpsol *)
let glucose = "glucose -verb=0 -model"
let glpsol = "glpsol --lp "

(* Arithmétique linéaire *)

(* module définissant les dictionnaires depuis les string vers ce qu'on veut *)
module SM = Map.Make(String)

(* une expression arithmétique linéaire est un dictionnaire 
   qui associe à des noms de variables des coefficient entiers,
   plus une constante entière *)
type poly = int SM.t * int

let merge_fun v c1 c2 = match c1, c2 with
    Some i1, Some i2 ->
      if i1 + i2 = 0 then None
      else Some (i1+i2)
  | None, Some c | Some c, None -> Some c
  | None, None -> None
  
let plus_poly : poly -> poly -> poly =
  fun (s, i) (t, j) ->
  SM.merge merge_fun s t, i + j

let inv_poly : poly -> poly =
  fun (s, i) ->
  SM.map (fun c -> -c) s, -i

let minus_poly : poly -> poly -> poly =
  fun s t ->
  plus_poly s (inv_poly t)


(* littéral en arithmétique linéaire *)
type lit =
    Leq of poly  (* Leq (p, c) équivaut à "p <= c" *)
  | Gr of poly   (* Gr (p, c)  équivaut à "p > c" *)

(* clause = liste de littéraux *)      
type clause = lit list

(* problème arithmétique = liste de clauses *)
type prob = clause list

(* modèle arithmétique = liste de littéraux 
   (ceux assignés à vrai dans le modèle) *)
type model = lit list
  
let distrib p1 p2 =
  List.fold_left
    (fun p c1 ->
      List.fold_left
        (fun p c2 -> (c1 @ c2)::p)
        p p2)
    [] p1
  

(* Logique propositionnelle *)

(* problème propositionnel = liste de clauses + infos
   clause = liste de litéraux
   litéraux = entier, positif si variable propositionnelle, 
   négatif si négation de variable *)
type sat_prob = {
  nb_clause : int;
  nb_var : int;
  clauses : int list list
}
  
type sat_model = int list


(* fonctions d'affichage *)  
let print_poly : out_channel -> int SM.t -> unit =
  fun ch s ->
  ignore @@ SM.fold (fun v i first ->
      if i > 0 && not first then
        Printf.fprintf ch " + %i %s" i v
      else
        Printf.fprintf ch " %i %s" i v;
      false
    ) s true
    
let print_lit : out_channel -> lit -> unit =
  fun ch -> function
  | Leq (s, i) ->
     print_poly ch s;
    Printf.fprintf ch " <= %i\n" i
  | Gr (s, i) ->
     print_poly ch s;
    Printf.fprintf ch " > %i\n" i

let print_prob : out_channel -> prob -> unit =
  fun ch p ->
  List.iter (fun c ->
    print_lit ch (List.hd c);
    List.iter (fun l -> output_string ch " | "; print_lit ch l) (List.tl c);
    output_string ch "\n") p
      
let print_model : out_channel -> model -> unit =
  fun ch m ->
  let merge_fun k a b =
    match a,b with
    | Some s, _ | _, Some s -> Some s
    | None, None -> None in
  let vars = List.fold_left
    (fun s -> function Leq(t,i)|Gr(t,i) -> SM.merge merge_fun s t) SM.empty m
  in
  let (v, _) = SM.choose vars in
  output_string ch ("maximize " ^ v ^ "\nsubject to\n");
  List.iter (print_lit ch) m;
  output_string ch "integer\n";
  SM.iter (fun v _ -> output_string ch (v ^ "\n")) vars;
  output_string ch "end\n"

let print_sat_clause : out_channel -> int list -> unit =
  fun ch c ->
  List.iter (Printf.fprintf ch "%i ") c;
  Printf.fprintf ch "0\n"
    
let print_sat_prob : out_channel -> sat_prob -> unit =
  fun ch p ->
  Printf.fprintf ch "p cnf %i %i\n" p.nb_var p.nb_clause;
  List.iter (print_sat_clause ch) p.clauses


    
exception Unsat
exception Sat
exception Model of string

let string_to_model : string -> sat_model =
  fun s ->
  let open Str in
  let intre = regexp "[+-]?[0-9]+" in
  let rec int_rec i =
    ignore (search_forward intre s i);
    let i = int_of_string (matched_string s) in
    if i = 0 then []
    else i :: int_rec (match_end ())
  in
   int_rec 0 

    
(* Appel externe à glucose
   retourne un modèle propositionnel si satisfiable
   lève Unsat si insatisfiable      *)
let call_glucose : sat_prob -> sat_model = fun p ->
  let from_g, to_g as proc = Unix.open_process glucose in
  print_sat_prob to_g p;
  flush to_g;
  close_out to_g;
  try
    while true do
      let l = input_line from_g in
      match l.[0] with
        's' ->
          if String.sub l 2 (String.length l - 2) <> "SATISFIABLE" then
            raise Unsat
      | 'v' -> raise (Model (String.sub l 2 (String.length l - 2)))
      | _ -> ()
    done;
    []
  with End_of_file -> raise Unsat
  | Model s ->
     string_to_model s
    
      
(* Appel externe à gplsol
   retourne true si satisfiable
   false si insatisfiable        *)
let call_glpsol : model -> bool = fun m ->
  let tmp_input = Filename.temp_file "smt" ".in" in
  let tmp_output = Filename.temp_file "smt" ".out" in
  let f = open_out tmp_input in
  print_model f m;
  close_out f;
  ignore (Sys.command (glpsol ^ " " ^ tmp_input ^ " -w " ^ tmp_output ^ " > /dev/null"));
  let f = open_in tmp_output in
  try while true do
    let l = input_line f in
    match l.[0] with
      's' ->
        Scanf.sscanf l "s mip %i %i %c %i"
          (fun _ _ c _ ->
              match c with
                'o' | 'f' -> raise Sat | 'n' | 'u' -> raise Unsat)
    | _ -> ()
    done; false
  with Sat -> close_in f; true | Unsat -> close_in f; false

    
(* module pour pouvoir hasher les expressions arithmétiques correctement *)    
module PolyHash = struct
  type t = poly

  let equal (s,i) (t,j) =
    i = j &&
    SM.compare compare s t = 0

  let hash (p, c) = let open Hashtbl in
    hash (SM.fold (fun s i h -> h + i * hash s) p (hash c))
end

module PH = Hashtbl.Make(PolyHash)

(* table de hachage des expressions arithmétiques vers les entiers 
   ajout d'une association entre expr et entier :
     PH.add poly_htbl expr entier
   recherche de l'entier associé à une expression expr :
     PH.find poly_htbl expr
   lève Not_found si rien associé
*)
let poly_htbl : int PH.t = PH.create 25

(* table de hachage des entiers vers les expressions arithmétiques
   ajout d'une association entre entier et expr :
     Hashtbl.add ylop_htbl entier expr
   recherche de l'entier associé à une expression expr :
     Hashtbl.find ylop_htbl entier
   lève Not_found si rien associé *)
let ylop_htbl : (int, poly) Hashtbl.t = Hashtbl.create 25
    

let model_to_clause : model -> clause =
  fun m -> 
  failwith "à compléter"
  
let arith_to_sat : prob -> sat_prob =
  fun p -> 
  failwith "à compléter"
    
let sat_to_arith : sat_model -> model =
  fun m ->
  failwith "à compléter"

let rec dpllt : prob -> bool =
  fun p ->
  failwith "à compléter"
